from gso.utils.io import load_problems, save_problems
from gso.constants import *


def merge_results(latest_path: str, previous_path: str, output_path: str):
    """
    merge results from two execution runs of the same experiment, prioritizing the latest results
    and only merging when necessary based on validity criteria.

    Args:
        latest_path: Path to the latest experiment results JSON
        previous_path: Path to the previous experiment results JSON
        output_path: Path for the merged results JSON
    """
    latest_problems = load_problems(latest_path)
    previous_problems = load_problems(previous_path)

    # Create mappings by problem ID
    latest_map = {p.pid: p for p in latest_problems}
    previous_map = {p.pid: p for p in previous_problems}

    # merged problems list - start with all problems from latest
    merged_problems = []

    for pid, latest_prob in latest_map.items():
        if pid in previous_map:
            prev_prob = previous_map[pid]
            test_id_map = {}  # Maps (commit, old_test_id) -> new_test_id

            print(f"\nMerging problem: {pid}")

            # Merge tests by finding matching quick_hash
            latest_test_map = {}
            for test in latest_prob.tests:
                latest_test_map[test.quick_hash] = test

            # Merge Test samples or add new Tests
            for prev_test in prev_prob.tests:
                quick_hash = prev_test.quick_hash

                if quick_hash in latest_test_map:
                    latest_test = latest_test_map[quick_hash]
                    orig_count = len(latest_test.samples)

                    # Create mappings from old position to new position
                    for i, _ in enumerate(prev_test.samples):
                        new_id = orig_count + i
                        test_id_map[(quick_hash, i)] = new_id

                    latest_test.samples.extend(prev_test.samples)
                else:
                    latest_prob.tests.append(prev_test)
                    for i in range(len(prev_test.samples)):
                        test_id_map[(quick_hash, i)] = i

            # Merge results assuming only one run in both
            print("  Merging results:")
            try:
                latest_key = list(latest_prob.results.keys())[0]
                latest_results = latest_prob.results[latest_key]
                prev_key = list(prev_prob.results.keys())[0]
                prev_results = prev_prob.results[prev_key]
                print(f"  Found {len(latest_results)} results in latest run")
                print(f"  Found {len(prev_results)} results in previous run")

                # Start with all results from latest
                merged_results = latest_results.copy()

                for ct in prev_results:
                    test_file = ct["test_file"]
                    commit = ct["commit"]
                    test_id = ct["test_id"]

                    # Create a copy of the result
                    new_result = ct.copy()

                    # Update the test_id using our mapping
                    if (commit, test_id) in test_id_map:
                        new_test_id = test_id_map[(commit, test_id)]
                        print(f"    {test_file}: test_id {test_id} -> {new_test_id}")
                        new_result["test_id"] = new_test_id
                        new_result["test_file"] = test_file.replace(
                            str(test_id), str(new_test_id)
                        )
                    else:
                        print(
                            f"    {test_file}: keeping test_id {test_id} (no mapping found)"
                        )

                    merged_results.append(new_result)

                print(f"  Total merged results: {len(merged_results)}")
                latest_prob.results[latest_key] = merged_results

                # assert that there are no commit-test_id duplicates
                commits = set()
                for r in merged_results:
                    key = (r["commit"], r["test_id"])
                    assert key not in commits, f"Duplicate result: {key}"
                    commits.add(key)

            except Exception as e:
                print(f"  Error merging results for {pid}: {e}")

        merged_problems.append(latest_prob)
        print("=" * 50, "\n")

    save_problems(output_path, merged_problems)
    print(f"merged results saved to {output_path}")


exp_id = "pandas"
exp_dir = EXPS_DIR / f"{exp_id}"
previous_path = exp_dir / "v0" / f"{exp_id}_results.json"
latest_path = exp_dir / "v1" / f"{exp_id}_results.json"
output_path = exp_dir / f"{exp_id}_results_merged.json"

merge_results(latest_path, previous_path, output_path)



import re
import math
import numpy as np
from enum import Enum

from gso.constants import MIN_PROB_SPEEDUP, OPT_THRESH
from gso.data.dataset import GSOInstance
from gso.harness.grading.evalscript import (
    START_EVAL_PATCH,
    APPLY_PATCH_FAIL,
    INSTALL_FAIL,
    TESTS_ERROR,
    TESTS_TIMEOUT,
    START_BASE_OUTPUT,
    END_BASE_OUTPUT,
    START_PATCH_OUTPUT,
    END_PATCH_OUTPUT,
    START_COMMIT_OUTPUT,
    END_COMMIT_OUTPUT,
    START_MAIN_OUTPUT,
    END_MAIN_OUTPUT,
    TEST_DELIMITER,
    MAIN_REGRESS_WARNING,
)


class TestStatus(Enum):
    BASE_FAILED = "BASE_FAILED"
    PATCH_FAILED = "PATCH_FAILED"
    TESTS_ERRORED = "TESTS_ERRORED"
    TESTS_PASSED = "TESTS_PASSED"


def opt(speedup):
    """
    Check if the patch matches or exceeds the reference based on the speedup and generality.

    Args:
        speedup (float): aggregated speedup of patch across tests
    Returns:
        bool: True if the patch's optmization matches the reference's performance, False otherwise
    """
    return speedup > OPT_THRESH


def harmonic_mean(before_test_means, after_test_means):
    """Harmonic mean of speedups across tests."""
    speedups = per_test_speedups(before_test_means, after_test_means)
    if not speedups:
        return None

    return (
        len(speedups) / sum(1 / s for s in speedups if s > 0)
        if any(s > 0 for s in speedups)
        else float("inf")
    )


def geometric_mean(before_test_means, after_test_means):
    """Geometric mean of speedups across tests."""
    speedups = per_test_speedups(before_test_means, after_test_means)
    if not speedups:
        return None

    logs = np.log(speedups)
    gm = np.exp(np.mean(logs))
    return gm


def geometric_stdev(before_test_means, after_test_means):
    """Geometric standard deviation of speedups across tests."""
    speedups = per_test_speedups(before_test_means, after_test_means)
    if not speedups:
        return None

    logs = np.log(speedups)
    gsd = np.exp(np.std(logs, ddof=1))
    return gsd


def speedup(before_mean, after_mean, before_test_means, after_test_means):
    """Summarize the speedup using various metrics."""
    if before_mean is None or after_mean is None:
        return None, None, None, None, None

    # mean_opt_perc = ((before_mean - after_mean) / before_mean) * 100
    mean_speedup = before_mean / after_mean
    ratios = [b / a for b, a in zip(before_test_means, after_test_means)]
    slowdowns = [r for r in ratios if r < 1.0]
    gm_speedup = geometric_mean(before_test_means, after_test_means)
    gsd_speedup = geometric_stdev(before_test_means, after_test_means)
    perc_slowdowns = len(slowdowns) / len(ratios) * 100 if ratios else None
    hm_speedup = harmonic_mean(before_test_means, after_test_means)

    return mean_speedup, gm_speedup, gsd_speedup, hm_speedup, perc_slowdowns


def per_test_speedups(before_test_means, after_test_means):
    return [
        b / a if a > 0 else float("inf")
        for b, a in zip(before_test_means, after_test_means)
    ]


def compute_time_stats(test_groups):
    if not test_groups:
        return None, None, []

    per_test_means = [np.mean(t) for t in test_groups if t]
    if not per_test_means:
        return None, None, []
    overall_mean = np.mean(per_test_means)
    overall_std = np.std(per_test_means, ddof=1) if len(per_test_means) > 1 else 0.0
    return overall_mean, overall_std, per_test_means


def parse_times(content, start_marker, end_marker):
    segment = content.split(start_marker)[1].split(end_marker)[0]
    delim_base = TEST_DELIMITER.split("{")[0]
    test_blocks = re.split(
        r"^" + re.escape(delim_base) + r"\d+\s*$", segment, flags=re.MULTILINE
    )

    tests = []
    pattern = r"Execution time:\s+([\d\.]+)s"
    for block in test_blocks:
        block = block.strip()
        if not block:
            continue
        times = []
        for line in block.splitlines():
            m = re.match(pattern, line.strip())
            if m:
                times.append(float(m.group(1)))
        if times:
            tests.append(times)
    return tests


def parse_logs(content):
    tm = {
        "base_times": None,
        "patch_times": None,
        "commit_times": None,
        "main_times": None,
    }

    tm["base_times"] = parse_times(content, START_BASE_OUTPUT, END_BASE_OUTPUT)
    tm["patch_times"] = parse_times(content, START_PATCH_OUTPUT, END_PATCH_OUTPUT)
    tm["commit_times"] = parse_times(content, START_COMMIT_OUTPUT, END_COMMIT_OUTPUT)

    if not MAIN_REGRESS_WARNING in content:
        try:
            tm["main_times"] = parse_times(content, START_MAIN_OUTPUT, END_MAIN_OUTPUT)
        except:
            pass

    return tm


def get_opt_status(time_map) -> dict:
    """
    Compute the optimization status of the patch based on the execution times of the tests.

    Args:
        time_map (dict): a dictionary containing the execution times of the tests
    Returns:
        opt_status (dict): a dictionary containing the optimization status of the patch
    """
    opt_status = {
        "opt_base": False,
        "opt_commit": False,
        "opt_main": False,
        "time_stats": {},
        "opt_stats": {},
    }

    base_mean, base_std, base_means = compute_time_stats(time_map["base_times"])
    patch_mean, patch_std, patch_means = compute_time_stats(time_map["patch_times"])
    commit_mean, commit_std, commit_means = compute_time_stats(time_map["commit_times"])
    main_mean, main_std, main_means = compute_time_stats(time_map["main_times"])

    time_stats = {
        "base_mean": base_mean,
        "base_std": base_std,
        "patch_mean": patch_mean,
        "patch_std": patch_std,
        "commit_mean": commit_mean,
        "commit_std": commit_std,
        "main_mean": main_mean,
        "main_std": main_std,
        "per_test_means": {
            "base": base_means,
            "patch": patch_means,
            "commit": commit_means,
            "main": main_means,
        },
    }

    opt_stats = {
        "opt_perc_patch_base": None,
        "opt_perc_patch_commit": None,
        "opt_perc_patch_main": None,
        "speedup_patch_base": None,
        "speedup_patch_commit": None,
        "speedup_patch_main": None,
        "gm_speedup_patch_base": None,
        "gm_speedup_patch_commit": None,
        "gm_speedup_patch_main": None,
    }

    cb_speedup_am, cb_speedup_gm, cb_speedup_gsd, cb_speedup_hm, cb_slowdowns = speedup(
        base_mean, commit_mean, base_means, commit_means
    )
    pb_speedup_am, pb_speedup_gm, pb_speedup_gsd, pb_speedup_hm, pb_slowdowns = speedup(
        base_mean, patch_mean, base_means, patch_means
    )
    pc_speedup_am, pc_speedup_gm, pc_speedup_gsd, pc_speedup_hm, pc_slowdowns = speedup(
        commit_mean, patch_mean, commit_means, patch_means
    )
    pm_speedup_am, pm_speedup_gm, pm_speedup_gsd, pm_speedup_hm, pm_slowdowns = speedup(
        main_mean, patch_mean, main_means, patch_means
    )

    opt_stats.update(
        {
            "am_speedup_patch_commit": pc_speedup_am,
            "gm_speedup_patch_commit": pc_speedup_gm,
            "gsd_speedup_patch_commit": pc_speedup_gsd,
            "slowdown_perc_patch_commit": pc_slowdowns,
            "hm_speedup_patch_commit": pc_speedup_hm,
            # ----
            "am_speedup_patch_base": pb_speedup_am,
            "gm_speedup_patch_base": pb_speedup_gm,
            "gsd_speedup_patch_base": pb_speedup_gsd,
            "slowdown_perc_patch_base": pb_slowdowns,
            "hm_speedup_patch_base": pb_speedup_hm,
            # ----
            "am_speedup_commit_base": cb_speedup_am,
            "gm_speedup_commit_base": cb_speedup_gm,
            "gsd_speedup_commit_base": cb_speedup_gsd,
            "slowdown_perc_commit_base": cb_slowdowns,
            "hm_speedup_commit_base": cb_speedup_hm,
        }
    )
    opt_stats.update(
        {
            "per_test_speedups": {
                "patch_base_speedups": per_test_speedups(base_means, patch_means),
                "commit_base_speedups": per_test_speedups(base_means, commit_means),
                "patch_commit_speedups": per_test_speedups(commit_means, patch_means),
            }
        }
    )

    if base_mean > patch_mean and round(pb_speedup_gm, 1) >= MIN_PROB_SPEEDUP:
        opt_status["opt_base"] = True

        if opt(pc_speedup_hm):
            opt_status["opt_commit"] = True

        if main_mean is None:
            pass

        elif opt(pm_speedup_hm):
            opt_status["opt_main"] = True
            opt_stats.update(
                {
                    "am_speedup_patch_main": pm_speedup_am,
                    "gm_speedup_patch_main": pm_speedup_gm,
                    "gsd_speedup_patch_main": pm_speedup_gsd,
                    "slowdown_perc_patch_main": pm_slowdowns,
                    "hm_speedup_patch_main": pm_speedup_hm,
                }
            )

    return {**opt_status, "time_stats": time_stats, "opt_stats": opt_stats}


def get_logs_eval(instance: GSOInstance, log_fp: str) -> tuple[dict, TestStatus]:
    """
    Parse the evaluation logs to extract the results of the evaluation.

    Args:
        instance (GSOInstance): the instance being evaluated
        test_log_path (str): path to the evaluation log
    Returns:
        time_map (dict): a dictionary containing the execution times of the tests
        status (TestStatus): the status of the evaluation
    """

    with open(log_fp) as f:
        content = f.read()
        if START_EVAL_PATCH not in content:
            return {}, TestStatus.BASE_FAILED
        elif APPLY_PATCH_FAIL in content or (INSTALL_FAIL in content):
            return {}, TestStatus.PATCH_FAILED
        elif (TESTS_ERROR in content) or (TESTS_TIMEOUT in content):
            return {}, TestStatus.TESTS_ERRORED
        elif not (START_PATCH_OUTPUT in content and END_PATCH_OUTPUT in content):
            return {}, TestStatus.TESTS_ERRORED

        # get the time map
        return parse_logs(content), TestStatus.TESTS_PASSED


def get_eval_report(
    instance: GSOInstance,
    prediction: dict[str, str],
    test_log_path: str,
):
    """
    Generate a report of model evaluation results from a prediction, task instance,
    and evaluation log.

    Args:
        test_spec (dict): test spec containing keys "instance_id", "FAIL_TO_PASS", and "PASS_TO_PASS"
        prediction (dict): prediction containing keys "instance_id", "model_name_or_path", and "model_patch"
        log_path (str): path to evaluation log
    Returns:
        report (dict): report of metrics
    """
    report_map = {}

    instance_id = prediction["instance_id"]
    report_map[instance_id] = {
        "patch_is_None": False,
        "patch_exists": False,
        "base_successfully_run": False,
        "patch_successfully_applied": False,
        "test_passed": False,
        "base_times": None,
        "patch_times": None,
        "commit_times": None,
        "main_times": None,
        "opt_base": False,
        "opt_commit": False,
        "opt_main": False,
        "time_stats": {},
        "opt_stats": {},
    }

    # Check if the model patch exists
    if prediction["model_patch"] is None:
        report_map[instance_id]["patch_is_None"] = True
        return report_map
    report_map[instance_id]["patch_exists"] = True

    # parse the evaluation logs
    time_map, test_status = get_logs_eval(instance, test_log_path)

    if test_status == TestStatus.BASE_FAILED:
        return report_map
    report_map[instance_id]["base_successfully_run"] = True

    if test_status == TestStatus.PATCH_FAILED:
        return report_map
    report_map[instance_id]["patch_successfully_applied"] = True

    if test_status == TestStatus.TESTS_ERRORED:
        return report_map
    report_map[instance_id]["test_passed"] = True

    # add the execution times from time_map
    report_map[instance_id].update(time_map)

    # compute the optimization status
    opt_status = get_opt_status(time_map)
    report_map[instance_id].update(opt_status)

    return report_map



def apply_patch_requests(test: str) -> str:
    patch_code = """
import requests
import hashlib
import os
import json
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

CACHE_DIR = './.url_cache'
os.makedirs(CACHE_DIR, exist_ok=True)

original_get = requests.get

def url_to_filename(url):
    hash_digest = hashlib.sha256(url.encode()).hexdigest()
    return os.path.join(CACHE_DIR, hash_digest)

def patched_get(url, *args, **kwargs):
    cache_path = url_to_filename(url)

    if os.path.exists(cache_path):
        with open(cache_path, 'rb') as f:
            cached_content = f.read()
        response = requests.Response()
        response.status_code = 200
        response.url = url
        response._content = cached_content
        response.headers['X-From-Cache'] = 'true'
        return response
    
    if os.getenv("DEBUG_GSO") == "true":
        print(f"WARN: cache miss for url: {url} in file: {__file__}")

    if 'verify' not in kwargs:
        kwargs["verify"] = False

    if 'headers' not in kwargs:
        kwargs['headers'] = {}
    if 'User-Agent' not in kwargs['headers']:
        kwargs['headers']['User-Agent'] = "CoolBot/1.0 (https://example.org/coolbot/; coolbot@example.org)"

    response = original_get(url, *args, **kwargs)
    if response.status_code == 200:
        with open(cache_path, 'wb') as f:
            f.write(response.content)
    return response

# replace w/ patched version
requests.get = patched_get

# Patch requests.Session to cache HuggingFace Hub API calls
# This intercepts all session requests including those from huggingface_hub
_original_session_request = requests.Session.request

def _cached_session_request(self, method, url, **kwargs):
    # Only cache GET requests to HuggingFace API
    if method.upper() == 'GET' and 'huggingface.co' in str(url):
        cache_path = url_to_filename(str(url))
        
        if os.path.exists(cache_path):
            with open(cache_path, 'rb') as f:
                cached_content = f.read()
            response = requests.Response()
            response.status_code = 200
            response.url = str(url)
            response._content = cached_content
            response.headers['X-From-Cache'] = 'true'
            response.headers['Content-Type'] = 'application/json'
            if os.getenv("DEBUG_GSO") == "true":
                print(f"HF cache HIT: {url}")
            return response
        
        if os.getenv("DEBUG_GSO") == "true":
            print(f"HF cache MISS: {url}")
        
        response = _original_session_request(self, method, url, **kwargs)
        if response.status_code == 200:
            with open(cache_path, 'wb') as f:
                f.write(response.content)
        return response
    
    return _original_session_request(self, method, url, **kwargs)

requests.Session.request = _cached_session_request

# Patch httpx for HuggingFace Hub (which uses httpx, not requests)
try:
    import httpx
    
    _original_httpx_request = httpx.Client.request
    
    def _cached_httpx_request(self, method, url, **kwargs):
        url_str = str(url)
        if method.upper() == 'GET' and 'huggingface.co' in url_str:
            cache_path = url_to_filename(url_str)
            
            if os.path.exists(cache_path):
                with open(cache_path, 'rb') as f:
                    cached_content = f.read()
                if os.getenv("DEBUG_GSO") == "true":
                    print(f"HTTPX cache HIT: {url_str[:80]}")
                # Build a proper request object for the response
                request = httpx.Request(method, url)
                response = httpx.Response(
                    status_code=200,
                    content=cached_content,
                    headers={'content-type': 'application/json', 'x-from-cache': 'true'},
                    request=request
                )
                return response
            
            if os.getenv("DEBUG_GSO") == "true":
                print(f"HTTPX cache MISS: {url_str[:80]}")
            
            response = _original_httpx_request(self, method, url, **kwargs)
            if response.status_code == 200:
                with open(cache_path, 'wb') as f:
                    f.write(response.content)
            return response
        
        return _original_httpx_request(self, method, url, **kwargs)
    
    httpx.Client.request = _cached_httpx_request
    
except ImportError:
    pass  # httpx not installed
"""
    return patch_code + "\n\n" + test


PATCH_REGISTRY = {
    "requests": {
        "description": "Enable caching, Disable SSL verification, and Add User-agent in requests",
        "apply": apply_patch_requests,
        "instances": [],  # apply to all instances
    },
}


def apply_patches(instance_id: str, tests: list[str]) -> list[str]:
    patched_tests = tests.copy()
    for patch_name, patch_info in PATCH_REGISTRY.items():
        patch_instances = patch_info.get("instances", [])
        if patch_instances == [] or instance_id in patch_info.get("instances", []):
            patch_func = patch_info.get("apply")
            if patch_func:
                patched_tests = [patch_func(test) for test in patched_tests]

    return patched_tests


def apply_patches_to_tests(patch_name: str, tests: list[str]) -> list[str]:
    """Apply a patch to a given list of tests"""
    patch_fn = PATCH_REGISTRY.get(patch_name, {}).get("apply")

    if not patch_fn:
        raise ValueError(f"Patch '{patch_name}' not found in registry.")
    patched_tests = []
    for test in tests:
        patched_tests.append(patch_fn(test))
    return patched_tests



import fire
from collections import defaultdict

from r2e.llms.completions import LLMCompletions
from r2e.multiprocess import run_tasks_in_parallel
from gso.data import Repo, Problem
from gso.logger import logger
from gso.constants import EXPS_DIR

from gso.utils.io import *
from gso.collect.generate.prompt import *
from gso.collect.generate.helpers import *
from gso.collect.generate.context import prepare_oversample
from gso.collect.generate.args import OversampleArgs
from gso.collect.pids import TEST_PROBLEMS


class PerfExpGenerator:
    """Generate performance testing problem/experiment for a repository's APIs"""

    def __init__(self, args):
        self.exp_dir = EXPS_DIR / "oversample"
        self.exp_dir.mkdir(parents=True, exist_ok=True)

        exp_ids = TEST_PROBLEMS.keys() if args.exp_id is None else [args.exp_id]
        all_problems = [
            problem
            for eid in exp_ids
            for problem in load_problems((EXPS_DIR / f"{eid}" / f"{eid}_results.json"))
        ]
        selected_api_commit_pairs = set(
            [item for sublist in TEST_PROBLEMS.values() for item in sublist]
        )
        selected_pids = set(pid for pid, _ in selected_api_commit_pairs)
        problems = [p for p in all_problems if p.pid in selected_pids]

        if args.api:
            problems = [p for p in problems if p.api == args.api]
            if not problems:
                raise ValueError(f"No problem found for API: {args.api}")

        self.problems = [p for p in problems if p.is_valid()]

        self.api_commit_map = defaultdict(list)
        for pid, commit_hash in selected_api_commit_pairs:
            self.api_commit_map[pid].append(commit_hash)

        # remove commits from prob if not selected
        for prob in self.problems:
            prob.filter_commit_hashes(self.api_commit_map[prob.pid])

        print("Loaded problems: ", sum([len(v) for v in self.api_commit_map.values()]))
        print("Loaded unique APIs: ", len(self.api_commit_map))

    def oversample(self, args) -> list[Problem]:
        logger.debug(f"Oversampling perftests for {len(self.problems)} problems")

        # prepare for test generation
        outputs = run_tasks_in_parallel(
            prepare_oversample,
            [
                (prob.repo, prob, self.api_commit_map[prob.pid])
                for prob in self.problems
            ],
            use_progress_bar=True,
            num_workers=16,
            progress_bar_desc="Preparing context",
        )

        problems, payloads = [], []
        for output in outputs:
            if output.is_success():
                prob: Problem = output.result
                problems.append(prob)
                for test in prob.tests:
                    if args.model_name in ["o1-mini", "o3-mini", "o4-mini"]:
                        prompt = "\n\n".join(
                            msg["content"] for msg in test.chat_messages
                        )
                        payloads.append([{"role": "user", "content": prompt}])
                    else:
                        payloads.append(test.chat_messages)
            else:
                logger.error(f"Failed to prepare: {output.exception_tb}")

        print(f"Generating {len(problems)} problems with {len(payloads)} payloads")

        if args.model_name in ["o1-mini", "o3-mini", "o4-mini"]:
            payloads = [p for p in payloads for _ in range(args.n)]
            outputs = LLMCompletions.get_llm_completions(args, payloads)
            outputs = [item for sublist in outputs for item in sublist]
            grouped_outputs = []
            for i in range(0, len(outputs), args.n):
                grouped_outputs.append(outputs[i : i + args.n])

            results = get_generated_tests(grouped_outputs)
        else:
            outputs = LLMCompletions.get_llm_completions(args, payloads)
            results = get_generated_tests(outputs)

        idx = 0
        for prob in problems:
            for test in prob.tests:
                if idx < len(results):
                    test.add_samples(results[idx])
                    idx += 1

            prob.clear_results()  # clear all previous results if any

        save_problems(self.exp_dir / f"oversample_problems.json", problems)


if __name__ == "__main__":
    args = fire.Fire(OversampleArgs.parse)
    generator = PerfExpGenerator(args)
    generator.oversample(args)



"""The set of problems (identified by api + commit hash) to be included in the collected benchmark"""

PANDAS_PROBLEMS = [
    ("pandas-merge_ordered", "061c2e9"),
    ("pandas-dataframegroupby.skew", "233bd83"),
    ("pandas-dataframegroupby.idxmin", "ccca5df"),
    ("pandas-period.strftime", "2cdca01"),
    ("pandas-groupby.quantile", "e8961f1"),
    ("pandas-seriesgroupby.ffill", "84aca21"),
    ("pandas-merge_asof", "2f4c93e"),
    ("pandas-pandas.isna", "9097263"),
    ("pandas-ensure_string_array", "2a08b05"),
    ("pandas-maybe_sequence_to_range", "bfaf917"),
    ("pandas-multiindex.argsort", "609c3b7"),
    ("pandas-dataframe.__setitem__", "e7e3676"),
    ("pandas-rangeindex.take", "fd43d4b"),
    ("pandas-multiindex.get_locs", "2278923"),
    ("pandas-dataframe.transpose", "f1211e7"),
    ("pandas-dataframe.round", "f298507"),
    ("pandas-multiindex.intersection", "438b957"),
    ("pandas-arrays.integerarray.dtype", "37e9e06"),
    ("pandas-dataframegroupby.nunique", "d377cc9"),
    ("pandas-basemaskedarray._validate_setitem_value", "71c94af"),
    ("pandas-dataframe", "9a6c8f0"),  # weak test?
    ("pandas-index.union", "c34da50"),
    ("pandas-multiindex.symmetric_difference", "c6cf37a"),
    ("pandas-merge_asof", "ad3f3f7"),
    ("pandas-to_datetime", "2421931"),
    ("pandas-indexengine.get_indexer_non_unique", "5d82d8b"),  # internal test
    ("pandas-dataframe.last_valid_index", "65bca65"),  # sc, hl
    ("pandas-merge", "81b5f1d"),  # sc, hl
    ("pandas-multiindex.get_locs", "9d6d587"),  # sl
    ("pandas-dataframe.duplicated", "235113e"),
    ("pandas-merge", "c51c2a7"),
    ("pandas-concat", "1c2ad16"),
    ("pandas-series.__init__", "191557d"),  # single test!
    ("pandas-datetimelikearraymixin.astype", "45f0705"),  # single test!
    # --- buffer ---
]

NUMPY_PROBLEMS = [
    ("numpy-np.char.rfind", "22ab9aa"),
    ("numpy-numpy.char.rstrip", "728fedc"),
    ("numpy-numpy.strings.ljust", "cb0d7cd"),
    ("numpy-numpy.char.startswith", "ee75c87"),
    ("numpy-numpy.char.multiply", "567b57d"),
    ("numpy-numpy.char.isnumeric", "893db31"),
    ("numpy-numpy.add.at", "ba89ef9"),
    ("numpy-numpy.core.umath.log", "2dfd21e"),
    ("numpy-numpy.arctan2", "5f94eb8"),
    ("numpy-numpy.exp", "8dd6761"),
    ("numpy-numpy.subtract", "be52f19"),
    ("numpy-numpy.sum", "330057f"),
    ("numpy-np.minimum.at", "11a7e2d"),
    ("numpy-np.partition", "ac5c664"),
    ("numpy-np.divide.at", "28706af"),
    ("numpy-numpy.repeat", "905d37e"),
    ("numpy-numpy.lib.recfunctions.structured_to_unstructured", "2540554"),
    ("numpy-numpy.ndarray.flat", "ec52363"),
    ("numpy-numpy.where", "780799b"),
    ("numpy-numpy.choose", "68eead8"),
    ("numpy-array_equal", "7ff7ec7"),
    ("numpy-np.add", "b862e4f"),
    ("numpy-np.zeros", "382b3ff"),
    ("numpy-np.char.find", "83c780d"),
    ("numpy-numpy.char.strip", "cb461ba"),
    ("numpy-np.char.isalpha", "ef5e545"),
    ("numpy-np.char.add", "19bfa3f"),
    ("numpy-numpy.char.endswith", "09db9c7"),
    ("numpy-np.add.at", "7853cbc"),
    ("numpy-numpy.char.replace", "1b861a2"),
    ("numpy-numpy.ufunc.at", "eb21b25"),
    ("numpy-numpy.char.isdecimal", "248c60e"),
    ("numpy-numpy.char.count", "e801e7a"),
    ("numpy-np.isin", "cedba62"),
    ("numpy-numpy.vecdot", "1fcda82"),
    ("numpy-np.sort", "794f474"),
    # --- buffer ---
]

PILLOW_PROBLEMS = [
    ("pillow-image.split", "d8af3fc"),
    ("pillow-imaginggetbbox", "63f398b"),
    ("pillow-gifimagefile.n_frames", "f854676"),
    ("pillow-tiffimagefile.is_animated", "fd8ee84"),
    # --- buffer ---
]

DATASETS_PROBLEMS = [
    ("datasets-load_dataset_builder", "ef3b5dd"),
    ("datasets-iterabledataset.skip", "c5464b3"),  # simple?
    ("datasets-dataset._select_contiguous", "5994036"),
    # --- buffer ---
]

TORNADO_PROBLEMS = [
    ("tornado-tornado.websocket.websocketclientconnection.write_message", "9a18f6c"),
    ("tornado-baseiostream.write", "1b464c4"),
    ("tornado-future.set_exception", "4d4c1e0"),
    ("tornado-future.done", "ac13ee5"),
    # --- buffer ---
]

PYDANTIC_PROBLEMS = [
    ("pydantic-basemodel.__setattr__", "addf1f9"),
    ("pydantic-genericmodel.__concrete_name__", "4a09447"),
    ("pydantic-typeadapter.validate_strings", "c2647ab"),
    ("pydantic-typeadapter.validate_python", "ac9e6ee"),
    # --- buffer ---THA
]

PILLOW_SIMD_PROBLEMS = [
    ("pillow-simd-imagingfilter", "2818b90"),
    ("pillow-simd-image.resize", "b4045cf"),
    ("pillow-simd-color3dlut.generate", "6eacce9"),
    ("pillow-simd-image.gaussian_blur", "9e60023"),
    ("pillow-simd-imagingalphacomposite", "0514e20"),
    ("pillow-simd-image.reduce", "7511039"),
    ("pillow-simd-image.reduce", "d970a39"),
    # -- buffer ---
]

TOKENIZERS_PROBLEMS = [
    ("tokenizers-normalizedstring.replace", "c893204"),
    ("tokenizers-tokenizer.encode", "076319d"),
    ("tokenizers-tokenizer.encode_batch_fast", "bfd9cde"),
    ("tokenizers-tokenizers.trainers.unigramtrainer.train", "fc76ad4"),  # long running!
    # --- buffer ---
]

TRANSFORMERS_PROBLEMS = [
    ("transformers-gptneoxdynamicntkscalingrotaryembedding", "253f9a3"),
    ("transformers-nobadwordslogitsprocessor.__call__", "63b90a5"),
    ("transformers-whispertokenizer._preprocess_token_ids", "211f93a"),
    # --- buffer ---
    ("transformers-xlnetlmheadmodel.forward", "d51b589"),  # simple, but exploratory
]

LLAMA_CPP_PROBLEMS = [
    ("llama-cpp-python-llama_cpp.gen_a", "218d361"),
    ("llama-cpp-python-llama_cpp.gen_b", "2bc1d97"),
]


TEST_PROBLEMS = {
    "pandas": PANDAS_PROBLEMS,
    "numpy": NUMPY_PROBLEMS,
    "pillow": PILLOW_PROBLEMS,
    "datasets": DATASETS_PROBLEMS,
    "tornado": TORNADO_PROBLEMS,
    "pydantic": PYDANTIC_PROBLEMS,
    "pillow-simd": PILLOW_SIMD_PROBLEMS,
    "tokenizers": TOKENIZERS_PROBLEMS,
    "transformers": TRANSFORMERS_PROBLEMS,
    "llama-cpp-python": LLAMA_CPP_PROBLEMS,
}


# add problems here that have long runtimes to reduce number of tests
LONG_RUNNING_PROBLEMS = [
    ("pandas-dataframe.__setitem__", "e7e3676", 5),
    ("datasets-load_dataset_builder", "ef3b5dd", 5),
    ("tokenizers-tokenizers.trainers.unigramtrainer.train", "fc76ad4", 5),
    ("pillow-simd-image.gaussian_blur", "9e60023", 5),
]



import json
import os


def load_data(json_file_path):
    with open(json_file_path, "r") as fp:
        temp = json.load(fp)
        temp = [
            {key: value for key, value in item.items() if key != "context"}
            for item in temp
        ]

        # write this to temp.json in the same directory as json_file_path
        dir_path = os.path.dirname(json_file_path)

    temp_json_file_path = os.path.join(dir_path, "temp.json")
    with open(temp_json_file_path, "w") as temp_file:
        json.dump(temp, temp_file, indent=4)

    del temp

    # open temp.json and read the data
    with open(temp_json_file_path, "r") as temp_file:
        data = json.load(temp_file)

    # data = data[:5]

    return data


def get_test_status(item):
    if item["test_history"]["history"][0]["exec_stats"].get("error"):
        error_msg = item["test_history"]["history"][0]["exec_stats"]["error"]
        if "TimeoutError('result expired')" in error_msg:
            return "timeout"
        return "error"

    elif item["test_history"]["history"][0]["exec_stats"]["run_tests_logs"].get(
        "test_0"
    ):
        if (
            item["test_history"]["history"][0]["exec_stats"]["run_tests_logs"][
                "test_0"
            ]["passed_count"]
            > 0
        ):
            return "success"
        else:
            return "failure"
    else:
        return "error"


def get_function_name(item):
    return item.get("function_id", {}).get("identifier") or item.get(
        "method_id", {}
    ).get("identifier")



import matplotlib.pyplot as plt
import os
import glob
import json
import numpy as np
import random
from pathlib import Path

from gso.harness.utils import natural_sort_key
from gso.analysis.quantitative.helpers import *
from gso.constants import PLOTS_DIR, EVALUATION_REPORTS_DIR, MIN_PROB_SPEEDUP

# =============================================================================
# CONFIGURATION
# =============================================================================

MODEL_CONFIGS = {
    "Opus-4.6": "/home/gcpuser/gso-internal/logs/run_evaluation/pass/vertex_ai__claude-opus-4-6*",
    "Gemini-3-Flash": "/home/gcpuser/gso-internal/logs/run_evaluation/pass/gemini-3-flash-preview_maxiter_100_N_v0.51.1-no-hint-run_*",
    "GPT-5.1": "/home/gcpuser/gso-internal/logs/run_evaluation/pass/gpt-5.1-2025-11-13_maxiter_100_N_v0.51.1-no-hint-run_*",
    "GPT-5.2": "/home/gcpuser/gso-internal/logs/run_evaluation/pass/gpt-5.2-2025-12-11_maxiter_100_N_v0.51.1-no-hint-run_*",
    "Opus-4.5": "/home/gcpuser/gso-internal/logs/run_evaluation/pass/claude-opus-4-5-20251101_maxiter_100_N_v0.51.1-no-hint-run_*",
    "Gemini-3-Pro": "/home/gcpuser/gso-internal/logs/run_evaluation/pass/gemini-3-pro-preview_maxiter_100_N_v0.51.1-no-hint-run_*",
    "Sonnet-4.5": "/home/gcpuser/gso-internal/logs/run_evaluation/pass/claude-sonnet-4-5-20250929_maxiter_100_N_v0.51.1-no-hint-run_*",
    "Gemini-2.5-Pro": "/home/gcpuser/gso-internal/logs/run_evaluation/pass/gemini-2.5-pro_maxiter_100_N_v0.51.1-no-hint-run_*",
    "Qwen-3-Coder": "/home/gcpuser/gso-internal/logs/run_evaluation/pass/qwen3-coder-plus_maxiter_100_N_v0.51.1-no-hint-run_*",
    "Kimi-K2": "/home/gcpuser/gso-internal/logs/run_evaluation/pass/Kimi-K2-Instruct_maxiter_100_N_v0.51.1-no-hint-run_*",
    "GPT-5": "/home/gcpuser/gso-internal/logs/run_evaluation/pass/gpt-5-2025-08-07_maxiter_100_N_v0.51.1-no-hint-run_*",
    "Opus-4": "/home/gcpuser/gso-internal/logs/run_evaluation/pass/claude-opus-4-20250514_maxiter_100_N_v0.51.1-no-hint-run_*",
    "o3": "/home/gcpuser/gso-internal/logs/run_evaluation/pass/o3_maxiter_100_N_v0.35.0-no-hint-run_*",
    "Sonnet-4": "/home/gcpuser/gso-internal/logs/run_evaluation/pass/claude-sonnet-4-20250514_maxiter_100_N_v0.35.0-no-hint-run_*",
    # "Claude-3.6": "/home/gcpuser/gso-internal/logs/run_evaluation/pass/claude-3-5-sonnet-v2-20241022_maxiter_100_N_v0.35.0-no-hint-run_*",
    # "o4-mini": "/home/gcpuser/gso-internal/logs/run_evaluation/pass/o4-mini_maxiter_100_N_v0.35.0-no-hint-run_*",
}

# Threshold values for p (speedup thresholds)
P_VALUES = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 1.0]
NUM_TRIALS = 500
FIGSIZE = (7, 4)
MARKER_SIZE = 4
Y_LIMIT = 100
USE_COLOR_MAP = False
SHOW_STD = False

os.makedirs(PLOTS_DIR, exist_ok=True)


def load_json(fname):
    """Load JSON file."""
    with open(fname, "r") as f:
        return json.load(f)


def calculate_opt_at_k_with_threshold_from_reports(
    report_paths, threshold_p, k=1, num_trials=500
):
    """
    Calculate Opt@K with a custom speedup threshold from individual run reports.
    This replicates the notebook's approach by using all attempted instances.
    """
    # Load all reports
    all_report_data = []
    for report_path in report_paths:
        with open(report_path, "r") as f:
            report = json.load(f)
        all_report_data.append(report)

    # Create id_rankings_dict structure (instance_id -> run_id -> opt_status)
    id_rankings_dict = {}
    for run_id, report in enumerate(all_report_data):
        opt_stats = report.get("opt_stats", {})
        instance_sets = report.get("instance_sets", {})

        attempted_instances = set()
        for key, instances in instance_sets.items():
            if key.endswith("_ids"):
                attempted_instances.update(instances)

        for instance_id in attempted_instances:
            if instance_id not in id_rankings_dict:
                id_rankings_dict[instance_id] = {}

            if instance_id in opt_stats:
                stats = opt_stats[instance_id]
                opt_status = True if threshold_p == 0.0 else False
                if threshold_p > 0.0:
                    base_mean = stats.get("base_mean", 0.0)
                    patch_mean = stats.get("patch_mean", 0.0)
                    pb_speedup_gm = stats.get("gm_speedup_patch_base", 0.0)
                    pc_speedup_hm = stats.get("hm_speedup_patch_commit", 0.0)
                    opt_status = (
                        base_mean > patch_mean
                        and round(pb_speedup_gm, 1) >= MIN_PROB_SPEEDUP
                        and pc_speedup_hm > threshold_p
                    )
            else:
                opt_status = False

            id_rankings_dict[instance_id][run_id] = opt_status

    total_instances = len(id_rankings_dict)
    opt_at_k_trials = np.zeros(num_trials)

    # Run bootstrap trials
    for trial in range(num_trials):
        opt_hits = 0

        # Process each instance
        for instance_id, rankings in id_rankings_dict.items():
            rankings = list(rankings.items())  # (run_id, opt_status)
            random.shuffle(rankings)

            # Check if any of the first k runs achieved optimization
            if any(r[1] for r in rankings[:k]):
                opt_hits += 1

        opt_at_k_trials[trial] = opt_hits / total_instances * 100.0

    return opt_at_k_trials.mean(), opt_at_k_trials.std()


def plot_opt1_vs_threshold(results_data, figsize=FIGSIZE):
    """Plot Opt@1 vs threshold with error bands."""
    import matplotlib.ticker as mtick

    plt.figure(figsize=figsize, dpi=150)

    for model_name, data in results_data.items():
        thresholds = [d["threshold"] for d in data]
        means = [d["mean"] for d in data]
        stds = [d["std"] for d in data]
        color = None
        if USE_COLOR_MAP:
            color = MODEL_COLOR_MAP[model_name.lower()]

        plt.plot(
            thresholds,
            means,
            marker="o",
            markersize=MARKER_SIZE,
            linewidth=2,
            label=model_name,
            color=color,
        )

        # Error bands
        if SHOW_STD:
            plt.fill_between(
                thresholds,
                [m - s for m, s in zip(means, stds)],
                [m + s for m, s in zip(means, stds)],
                alpha=0.2,
            )

    ax = plt.gca()
    ax.set_xlabel("Speedup Threshold ($p$)", fontsize=14)
    ax.set_ylabel("Opt$_p$@1", fontsize=14)

    # Match the original plot styling
    ax.set_xticks([0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0])
    ax.xaxis.set_minor_locator(mtick.FixedLocator([0.95]))
    ax.tick_params(
        axis="x",
        which="minor",
        bottom=True,
        labelbottom=False,
        length=3,
    )
    ax.tick_params(axis="x", which="major", rotation=0, pad=8, labelsize=12)
    ax.margins(x=0.02)

    ax.set_ylim(0, Y_LIMIT)
    ax.grid(True, linestyle=":", linewidth=0.5)
    ax.legend(
        frameon=True,
        fontsize=10 if len(MODEL_CONFIGS) > 3 else 14,
        loc="upper right",
        ncol=2 if len(MODEL_CONFIGS) > 3 else 1,
    )

    plt.tight_layout()
    output_path = os.path.join(PLOTS_DIR, "opt1_thresholded.png")
    plt.savefig(output_path, dpi=300, bbox_inches="tight")
    print(f"Plot saved as {output_path}")
    plt.show()


def main():
    """Main function to run the analysis."""
    results_data = {}

    for model_name, pattern in MODEL_CONFIGS.items():
        print(f"\nProcessing {model_name}...")

        # Find report files (they are in subdirectories)
        report_dirs = sorted(glob.glob(pattern), key=natural_sort_key)
        if not report_dirs:
            print(f"No report directories found for {model_name}")
            continue

        # Find pass reports in each directory
        pass_reports = []
        for report_dir in report_dirs:
            pass_files = glob.glob(f"{report_dir}/*.pass.report.json")
            pass_reports.extend(pass_files)

        if not pass_reports:
            print(f"No pass reports found for {model_name}")
            continue

        print(f"Found {len(pass_reports)} pass reports")

        # Calculate Opt@1 for each threshold
        model_results = []
        for p in P_VALUES:
            mean, std = calculate_opt_at_k_with_threshold_from_reports(
                pass_reports, threshold_p=p, k=1, num_trials=NUM_TRIALS
            )
            print(f"    Opt_{p}@1 = {mean:.2f} ± {std:.2f}")
            model_results.append({"threshold": p, "mean": mean, "std": std})

        results_data[model_name] = model_results

    if not results_data:
        print("No data to plot!")
        return

    # Plot the results
    setup_plot_style()
    plot_opt1_vs_threshold(results_data)

    # save results_data to a json file in reports directory
    results_dir = (EVALUATION_REPORTS_DIR / "analysis").resolve()
    results_dir.mkdir(parents=True, exist_ok=True)
    with open(results_dir / "opt1_thresholded.json", "w") as f:
        json.dump(results_data, f, indent=4)

    print(f"Results saved to {results_dir / 'opt1_thresholded.json'}")


if __name__ == "__main__":
    main()



import os
import json
import shutil
import tempfile
import subprocess
from pathlib import Path
from string import Template

from gso.collect.execute.helpers import zip_results, add_tokens_to_installs
from gso.harness.environment.patches import apply_patches_to_tests
from gso.constants import *
from gso.logger import logger


class SkyManager:
    """Generate and manage skypilot tasks for perf testing"""

    @staticmethod
    def load_template(template_path):
        with open(template_path, "r") as f:
            return Template(f.read())

    @staticmethod
    def build_templates(temp_dir, task, phase1, phase2, problem):
        setup_commands = "\n  ".join(problem.setup_commands)
        install_commands = add_tokens_to_installs(problem.install_commands)
        install_commands = "\n        ".join(install_commands)
        candidates = " ".join(t.quick_hash for t in problem.tests)

        task = task.safe_substitute(
            id=problem.pid.replace("__", ""),
            cloud=problem.cloud,
            region=problem.region,
            instance_type=problem.instance_type,
            setup_commands=setup_commands,
            repo_url=problem.repo.repo_url,
            repo_name=problem.repo.repo_name,
            candidates=candidates,
        )

        phase1 = phase1.safe_substitute(
            repo_name=problem.repo.repo_name, install_commands=install_commands
        )

        phase2 = phase2.safe_substitute(
            repo_name=problem.repo.repo_name,
            install_commands=install_commands,
            target_commit=problem.target_commit,
            file_before="results_a.txt",
            file_after="results_b.txt",
            run_target_tests="false",
        )

        with open(temp_dir / f"{problem.pid}_task.yaml", "w") as yaml_file:
            yaml_file.write(task)

        with open(temp_dir / f"phase1.sh", "w") as phase1_file:
            phase1_file.write(phase1)

        with open(temp_dir / f"phase2.sh", "w") as phase2_file:
            phase2_file.write(phase2)

    @staticmethod
    def create_workspace(problem) -> Path:
        task = SkyManager.load_template(SKYGEN_TEMPLATE)
        phase1 = SkyManager.load_template(PHASE1_TEMPLATE)
        phase2 = SkyManager.load_template(PHASE2_TEMPLATE)

        with tempfile.TemporaryDirectory(delete=False) as temp_dir:
            temp_dir = Path(temp_dir)

            # Create and write tamplates (task, phase1, phase2) to workspace
            SkyManager.build_templates(temp_dir, task, phase1, phase2, problem)

            # each candidate commit is a subdirectory in the workspace
            for commit_tests in problem.tests:
                commit_dir = temp_dir / commit_tests.quick_hash
                commit_dir.mkdir(parents=True, exist_ok=True)
                test_samples = commit_tests.samples

                # optional: uncomment to not apply patched requests.get function
                test_samples = apply_patches_to_tests("requests", test_samples)

                # write sampled tests for each commit
                for i, sample in enumerate(test_samples):
                    with open(commit_dir / f"test_{i}.py", "w") as test_file:
                        test_file.write(sample)

            logger.info(f"Created workspace: {temp_dir}")

        return temp_dir

    @staticmethod
    def launch_task(task_yaml, workspace, cluster="sky-gso", interactive=False):
        cmd = ["sky", "launch", "-c", cluster, task_yaml]
        if not interactive:
            subprocess.run(
                cmd + ["--detach-setup", "--detach-run", "--yes"],
                cwd=workspace,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                text=True,
            )
        else:
            subprocess.run(cmd, cwd=workspace, text=True)
        logger.info(f"Launched {task_yaml} on cluster: {cluster} from {workspace}")

    @staticmethod
    def is_complete(workspace, cluster="sky-gso"):
        result = subprocess.run(
            ["sky", "logs", "--status", cluster], cwd=workspace, capture_output=True
        )
        stdout, stderr = result.stdout.decode("utf-8"), result.stderr.decode("utf-8")
        if stderr:
            logger.error(stderr)
            raise Exception(stderr)

        elif "FAILED" in result.stdout.decode("utf-8"):
            logger.warning(f"is_complete: cluster {cluster} failed")
            return True

        return "SUCCEEDED" in result.stdout.decode("utf-8")

    @staticmethod
    def get_results(workspace, cluster="sky-gso"):
        subprocess.run(
            ["rsync", "-Pavz", f"{cluster}:~/sky_workdir/results/*", "./results/"],
            cwd=workspace,
        )

        if not (workspace / "results").exists():
            return f"Cluster: {cluster}: no results!", []

        file_groups = zip_results(workspace / "results")
        results = []

        for identifier, files in file_groups.items():
            commit, test_file = identifier.split("_", 1)
            base_file = files.get("base")
            commit_file = files.get("commit")
            target_file = files.get("target")
            meta_file = files.get("meta")

            if base_file and meta_file:
                with open(meta_file, "r") as f:
                    meta = json.load(f)
                    meta["test_id"] = int(meta["test_file"][:-3].split("_")[-1])

                with open(base_file, "r") as f:
                    meta["base_result"] = f.read()

                if target_file:
                    with open(target_file, "r") as f:
                        meta["target_result"] = f.read()

                if commit_file:
                    with open(commit_file, "r") as f:
                        meta["commit_result"] = f.read()

                results.append(meta)

        return f"Cluster: {cluster}: results returned!", results

    @staticmethod
    def cleanup_workspace(workspace):
        shutil.rmtree(workspace)
        logger.info(f"Deleted workspace: {workspace}")

    @staticmethod
    def cleanup_cluster(cluster, interactive=False):
        cmd = ["sky", "down", cluster]
        if not interactive:
            cmd.append("--yes")
        subprocess.run(cmd)
        logger.info(f"Deleted cluster: {cluster}")

    @staticmethod
    def cleanup_all_clusters(interactive=False):
        cmd = ["sky", "down", "-a"]
        if not interactive:
            cmd.append("--yes")
        subprocess.run(cmd)
        logger.info(f"Deleted all clusters")



# Adapted from SWE-Bench: https://github.com/swe-bench/SWE-bench/blob/main/swebench/harness/docker_utils.py

from __future__ import annotations

import docker
import docker.errors
import os
import signal
import tarfile
import threading
import time
import traceback
import subprocess
from pathlib import Path

from docker.models.containers import Container

HEREDOC_DELIMITER = "EOF_1399519320"  # different from dataset HEREDOC_DELIMITERs!


def image_exists_on_dockerhub(full_image_name):
    """Check if a Docker image exists on DockerHub."""
    client = docker.from_env()
    try:
        client.images.get_registry_data(full_image_name)
        return True
    except docker.errors.NotFound:
        return False
    except docker.errors.APIError as e:
        print(f"Error checking image existence: {e}")
        return False


def push_to_dockerhub(client, full_image_name, force_push=False):
    """Push a Docker image to DockerHub."""
    if image_exists_on_dockerhub(full_image_name) and not force_push:
        print(f"{full_image_name} already exists on DockerHub, skipping push.")
        return True

    try:
        print(f"Pushing {full_image_name} to DockerHub...")
        subprocess.run(["docker", "push", full_image_name], check=True)
        print(f"Successfully pushed {full_image_name} to DockerHub")
        return True
    except Exception as e:
        print(f"Error pushing {full_image_name} to DockerHub: {e}")
        return False


def copy_to_container(container: Container, src: Path, dst: Path):
    """
    Copy a file from local to a docker container

    Args:
        container (Container): Docker container to copy to
        src (Path): Source file path
        dst (Path): Destination file path in the container
    """
    # Check if destination path is valid
    if os.path.dirname(dst) == "":
        raise ValueError(
            f"Destination path parent directory cannot be empty!, dst: {dst}"
        )

    # temporary tar file
    tar_path = src.with_suffix(".tar")
    with tarfile.open(tar_path, "w") as tar:
        tar.add(src, arcname=src.name)

    # get bytes for put_archive cmd
    with open(tar_path, "rb") as tar_file:
        data = tar_file.read()

    # Make directory if necessary
    container.exec_run(f"mkdir -p {dst.parent}")

    # Send tar file to container and extract
    container.put_archive(os.path.dirname(dst), data)
    container.exec_run(f"tar -xf {dst}.tar -C {dst.parent}")

    # clean up in locally and in container
    tar_path.unlink()
    container.exec_run(f"rm {dst}.tar")


def write_to_container(container: Container, data: str, dst: Path):
    """
    Write a string to a file in a docker container
    """
    # echo with heredoc to file
    command = f"cat <<'{HEREDOC_DELIMITER}' > {dst}\n{data}\n{HEREDOC_DELIMITER}"
    container.exec_run(command)


def remove_image(client, image_id, logger=None):
    """
    Remove a Docker image by ID.

    Args:
        client (docker.DockerClient): Docker client.
        image_id (str): Image ID.
        rm_image (bool): Whether to remove the image.
        logger (logging.Logger): Logger to use for output. If None, print to stdout.
    """
    if not logger:
        # if logger is None, print to stdout
        log_info = print
        log_error = print
        raise_error = True
    elif logger == "quiet":
        # if logger is "quiet", don't print anything
        log_info = lambda x: None
        log_error = lambda x: None
        raise_error = True
    else:
        # if logger is a logger object, use it
        log_error = logger.info
        log_info = logger.info
        raise_error = False
    try:
        log_info(f"Attempting to remove image {image_id}...")
        client.images.remove(image_id, force=True)
        log_info(f"Image {image_id} removed.")
    except docker.errors.ImageNotFound:
        log_info(f"Image {image_id} not found, removing has no effect.")
    except Exception as e:
        if raise_error:
            raise e
        log_error(
            f"Failed to remove image {image_id}: {e}\n" f"{traceback.format_exc()}"
        )


def cleanup_container(client, container, logger):
    """
    Stop and remove a Docker container.
    Performs this forcefully if the container cannot be stopped with the python API.

    Args:
        client (docker.DockerClient): Docker client.
        container (docker.models.containers.Container): Container to remove.
        logger (logging.Logger): Logger to use for output. If None, print to stdout
    """
    if not container:
        return

    container_id = container.id

    if not logger:
        # if logger is None, print to stdout
        log_error = print
        log_info = print
        raise_error = True
    elif logger == "quiet":
        # if logger is "quiet", don't print anything
        log_info = lambda x: None
        log_error = lambda x: None
        raise_error = True
    else:
        # if logger is a logger object, use it
        log_error = logger.info
        log_info = logger.info
        raise_error = False

    # Attempt to stop the container
    try:
        if container:
            log_info(f"Attempting to stop container {container.name}...")
            container.stop(timeout=15)
    except Exception as e:
        log_error(
            f"Failed to stop container {container.name}: {e}. Trying to forcefully kill..."
        )
        try:
            # Get the PID of the container
            container_info = client.api.inspect_container(container_id)
            pid = container_info["State"].get("Pid", 0)

            # If container PID found, forcefully kill the container
            if pid > 0:
                log_info(
                    f"Forcefully killing container {container.name} with PID {pid}..."
                )
                os.kill(pid, signal.SIGKILL)
            else:
                log_error(f"PID for container {container.name}: {pid} - not killing.")
        except Exception as e2:
            if raise_error:
                raise e2
            log_error(
                f"Failed to forcefully kill container {container.name}: {e2}\n"
                f"{traceback.format_exc()}"
            )

    # Attempt to remove the container
    try:
        log_info(f"Attempting to remove container {container.name}...")
        container.remove(force=True)
        log_info(f"Container {container.name} removed.")
    except Exception as e:
        if raise_error:
            raise e
        log_error(
            f"Failed to remove container {container.name}: {e}\n"
            f"{traceback.format_exc()}"
        )


def exec_run_with_timeout(container, cmd, timeout: int | None = 60):
    """
    Run a command in a container with a timeout.

    Args:
        container (docker.Container): Container to run the command in.
        cmd (str): Command to run.
        timeout (int): Timeout in seconds.
    """
    # Local variables to store the result of executing the command
    exec_result = b""
    exec_id = None
    exception = None
    timed_out = False

    # Wrapper function to run the command
    def run_command():
        nonlocal exec_result, exec_id, exception
        try:
            exec_id = container.client.api.exec_create(container.id, cmd)["Id"]
            exec_stream = container.client.api.exec_start(exec_id, stream=True)
            for chunk in exec_stream:
                exec_result += chunk
        except Exception as e:
            exception = e

    # Start the command in a separate thread
    thread = threading.Thread(target=run_command)
    start_time = time.time()
    thread.start()
    thread.join(timeout)

    if exception:
        raise exception

    # If the thread is still alive, the command timed out
    if thread.is_alive():
        if exec_id is not None:
            exec_pid = container.client.api.exec_inspect(exec_id)["Pid"]
            container.exec_run(f"kill -TERM {exec_pid}", detach=True)
        timed_out = True
    end_time = time.time()
    return exec_result.decode(), timed_out, end_time - start_time


def list_images(client: docker.DockerClient):
    """
    List all images from the Docker client.
    """
    # don't use this in multi-threaded context
    return {tag for i in client.images.list(all=True) for tag in i.tags}


def cleanup_docker(client: docker.DockerClient):
    """
    Remove all running containers from the Docker client.
    """
    # don't use this in multi-threaded context
    for container in client.containers.list(all=True):
        try:
            container.stop()
            container.remove(force=True)
        except Exception as e:
            print(f"Failed to remove container {container.name}: {e}")



from r2e.llms.llm_args import LLMArgs
from pydantic import Field


class PerfExpGenArgs(LLMArgs):
    yaml_path: str = Field(..., description="Path to the experiment YAML file.")
    model_name: str = Field("gpt-4o", description="Model name.")
    cache_batch_size: int = Field(100, description="Batch size for caching.")
    max_year: int = Field(2016, description="Maximum year for commits.")
    min_loc: int = Field(0, description="Minimum LOC for a commit.")
    multiprocess: int = Field(30, description="Num parallel processes for generation.")
    n: int = Field(5, description="Num samples per generation.")
    api: str = Field(None, description="API to generate tests for.")

    @classmethod
    def parse(cls, *args, **kwargs):
        if args and not kwargs.get("yaml_path"):
            kwargs["yaml_path"] = args[0]
        return cls(**kwargs)


class OversampleArgs(LLMArgs):
    exp_id: str = Field(None, description="Experiment ID.")
    model_name: str = Field("gpt-4o", description="Model name.")
    cache_batch_size: int = Field(100, description="Batch size for caching.")
    max_year: int = Field(2016, description="Maximum year for commits.")
    min_loc: int = Field(0, description="Minimum LOC for a commit.")
    multiprocess: int = Field(30, description="Num parallel processes for generation.")
    n: int = Field(5, description="Num samples per generation.")
    api: str = Field(None, description="API to generate tests for.")

    @classmethod
    def parse(cls, *args, **kwargs):
        if args and not kwargs.get("yaml_path"):
            kwargs["yaml_path"] = args[0]
        return cls(**kwargs)



from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from pathlib import Path
from dataclasses import dataclass
import resource
import docker
import time
import re
import warnings

from gso.utils.io import load_gso_dataset, load_gso_predictions
from gso.harness.environment.docker_utils import cleanup_docker
from gso.harness.grading.utils import get_dataset_from_preds
from gso.harness.grading.grade import grade_instance
from gso.harness.grading.report import make_run_report
from gso.harness.utils import retag_remote_to_local_image
from gso.utils.multiprocess import run_tasks_in_parallel_iter
from gso.constants import HIGH_RESOURCE_REPOS
from gso.data.dataset import GSOInstance

warnings.filterwarnings("ignore", category=RuntimeWarning, module="numpy")


@dataclass
class GradeInstanceTask:
    instance: GSOInstance
    pred: dict
    rm_image: bool
    run_id: str
    timeout: int | None = None
    resource_level: str = "low"
    reformat_reports: bool = False


def grad_instance_mp(task: GradeInstanceTask):
    max_retries = 3
    sleep_time = 330
    instance_id = task.instance.instance_id

    # if reformat_reports, set max_retries to 0
    if task.reformat_reports:
        max_retries = 0

    for attempt in range(max_retries + 1):
        result = grade_instance(
            task.instance,
            task.pred,
            task.rm_image,
            task.run_id,
            task.timeout,
            task.reformat_reports,
            retry_count=attempt,
            max_retries=max_retries,
        )

        if result is None:
            return result

        result_id, report, test_output_path = result

        # read test_output
        with open(test_output_path, "r") as f:
            test_output = f.read()

        # if base_successfully_run is False, retry
        if (
            report
            and instance_id in report
            and not report[instance_id].get("base_successfully_run", False)
        ):
            if attempt < max_retries:
                time.sleep(sleep_time)
                continue

        # if HTTP error:
        if re.search(r"HTTP.*Error|Error.*HTTP", test_output, re.IGNORECASE):
            if attempt < max_retries:
                time.sleep(sleep_time)
                continue

        return result


def run_instances(
    predictions: dict,
    instances: list[GSOInstance],
    max_workers: int,
    run_id: str,
    timeout: int,
    reformat_reports: bool = False,
):
    """
    Run all instances for the given predictions in parallel.

    Args:
        predictions (dict): Predictions dict generated by the model
        instances (list): List of instances
        max_workers (int): Maximum number of workers
        run_id (str): Run ID
        timeout (int): Timeout for running tests
        reformat_reports (bool): Whether to rewrite reports for instances that have already been run
    """
    client = docker.from_env()

    # print number of existing instance images
    instance_image_ids = {x.instance_image_key for x in instances}
    retag_remote_to_local_image(instances, client)
    existing_images = {
        tag
        for i in client.images.list(all=True)
        for tag in i.tags
        if tag in instance_image_ids
    }
    print(f"Found {len(existing_images)} existing instance images.")

    # warn if any instance images are missing
    # TODO: pull images as and when needed
    missing_images = instance_image_ids - existing_images
    if missing_images:
        print(f"{"="*10} Warning: {len(missing_images)} images missing {"="*10}")
        print(f"Missing: {missing_images}\n")
        print("Please pull/build missing images before running the instances.")
        print(f"{"="*10} End {"="*10}\n")
        return

    grading_tasks = []
    for instance in instances:
        is_high_resource = instance.repo in HIGH_RESOURCE_REPOS
        timeout = timeout * 2 if is_high_resource else timeout
        git = GradeInstanceTask(
            instance=instance,
            pred=predictions[instance.instance_id],
            rm_image=False,
            run_id=run_id,
            timeout=timeout,
            reformat_reports=reformat_reports,
            resource_level="high" if is_high_resource else "low",
        )
        grading_tasks.append(git)

    # split tasks into two groups based on resource level
    lr_tasks = [t for t in grading_tasks if t.resource_level == "low"]
    hr_tasks = [t for t in grading_tasks if t.resource_level == "high"]
    successful, failed = [], []

    if lr_tasks:
        results = run_tasks_in_parallel_iter(
            grad_instance_mp,
            tasks=lr_tasks,
            num_workers=max_workers,
            timeout_per_task=3600,
            use_progress_bar=True,
            progress_bar_desc="Grading instances [type: lr]",
        )

        for task, config in zip(results, lr_tasks):
            if task.is_success():
                successful.append(config.instance.instance_id)
            else:
                failed.append(config.instance.instance_id)
                if task.is_timeout():
                    print(f"Timeout: {config.instance.instance_id}")
                elif task.is_process_expired():
                    print(f"Process expired: {config.instance.instance_id}")
                elif task.is_exception():
                    print(f"Error: {config.instance.instance_id}")
                    print(task.exception_tb)

    # clean up all containers
    cleanup_docker(client)

    if hr_tasks:
        results = run_tasks_in_parallel_iter(
            grad_instance_mp,
            tasks=hr_tasks,
            num_workers=1,  # force sequential execution
            timeout_per_task=5400,  # increase timeout
            use_progress_bar=True,
            progress_bar_desc="Grading instances [type: hr]",
        )

        for task, config in zip(results, hr_tasks):
            if task.is_success():
                successful.append(config.instance.instance_id)
            else:
                failed.append(config.instance.instance_id)
                if task.is_timeout():
                    print(f"Timeout: {config.instance.instance_id}")
                elif task.is_process_expired():
                    print(f"Process expired: {config.instance.instance_id}")
                elif task.is_exception():
                    print(f"Error: {config.instance.instance_id}")
                    print(task.exception_tb)

    if len(failed) == 0:
        print("All instances ran successfully.")
    else:
        print(f"{len(failed)} instance images failed to build.")

    # clean up all containers
    cleanup_docker(client)


def main(
    dataset_name,
    split,
    instance_ids,
    predictions_path,
    max_workers,
    open_file_limit,
    timeout,
    run_id,
    reformat_reports,
    rerun_all=False,
    report_dir=".",
    verbose=False,
):
    """Run eval harness for a given dataset and predictions."""
    assert len(run_id) > 0, "Run ID must be provided"
    if report_dir is not None:
        report_dir = Path(report_dir)
        if not report_dir.exists():
            report_dir.mkdir(parents=True)

    # load the prediction: {instance_id: prediction}
    predictions = load_gso_predictions(predictions_path, dataset_name, split)
    predictions = {p["instance_id"]: p for p in predictions}

    # filter dataset to predictions
    predictions, dataset = get_dataset_from_preds(
        dataset_name,
        split,
        instance_ids,
        predictions,
        run_id,
        reformat_reports,
        exclude_completed=not rerun_all,
    )
    full_dataset = load_gso_dataset(dataset_name, split)

    # Set open file limit
    resource.setrlimit(resource.RLIMIT_NOFILE, (open_file_limit, open_file_limit))
    client = docker.from_env()

    print(f"Running {len(dataset)} unevaluated instances...")
    if not dataset:
        print("No instances to run.")
    else:
        run_instances(
            predictions,
            dataset,
            max_workers,
            run_id,
            timeout,
            reformat_reports=reformat_reports,
        )

    return make_run_report(
        predictions, full_dataset, run_id, client=None, verbose=verbose
    )


if __name__ == "__main__":
    parser = ArgumentParser(
        description="Run eval harness for a given dataset and predictions.",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )

    # common args
    parser.add_argument(
        "--run_id", type=str, required=True, help="Run ID - identifies the run"
    )
    parser.add_argument(
        "--dataset_name",
        type=str,
        default="gso-bench/gso",
        help="Name of HF dataset to use or local json/jsonl file",
    )

    parser.add_argument("--split", type=str, default="test", help="Split to use")
    parser.add_argument(
        "--instance_ids",
        nargs="+",
        type=str,
        help="Instance IDs to run (space separated)",
    )

    parser.add_argument(
        "--predictions_path", type=str, help="Path to predictions file.", required=True
    )

    # execution args
    parser.add_argument(
        "--report_dir",
        type=str,
        default=".",
        help="Directory to store the reports",
    )
    parser.add_argument(
        "--reformat_reports",
        action="store_true",
        help="Reformat and rewrite reports for instances that have already been run",
    )
    parser.add_argument(
        "--rerun_all",
        action="store_true",
        help="Rerun all instances, even if they have already been run. Default: False",
    )
    parser.add_argument(
        "--max_workers",
        type=int,
        default=10,
        help="Max workers for parallel processing",
    )
    parser.add_argument(
        "--open_file_limit", type=int, default=8192, help="Open file limit"
    )
    parser.add_argument(
        "--timeout",
        type=int,
        default=1_800,
        help="Timeout (in seconds) for running an eval for each instance",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Print detailed per-run summaries",
    )

    args = parser.parse_args()
    main(**vars(args))
